/*
 * puzzle.c
 *
 *
 * reconstruction of tree topologies with quartet puzzling
 *
 * by Korbinian Strimmer and Arndt von Haeseler (1995)
 * this file belongs to the Puzzle version 2.3 distribution
 *
 */


#define EXTERN
#define TRUE 1
#define FALSE 0

#define VERSION    "2.3"
#define DATE       "1 December 1995"

#define INFILE     "infile"
#define OUTFILE    "puzzle.out"
#define TREEFILE   "qptrees"

/* prototypes */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <ctype.h>
#include "matrixut.h"
#include "qml.h"

/* abbreviation */
typedef unsigned long int ulint;

/* tree structure */
typedef struct oneedge {
	/* pointer to other three edges */
	struct oneedge *up;
	struct oneedge *downleft;
	struct oneedge *downright;
	int numedge;     /* number of edge */
	ulint edgeinfo;  /* value of this edge */
	int *edgemap;    /* pointer to the local edgemap */
} ONEEDGE;

/* variables */
cmatrix biparts;      /* bipartitions of current quartet puzzling tree */
double tstvratio;     /* expected transition/transversion ratio */
FILE *ofp;            /* pointer to output file */
FILE *seqfp;          /* pointer to sequence input file */
FILE *tfp;            /* pointer to tree file */
int *edgeofleaf;      /* vector with edge number of all leaves */
int Approx_optn;      /* approximate ml */
int chooseA;          /* leaf variable */
int chooseB;          /* leaf variable */
int codelength;       /* how many characters define one state */
int column;           /* used for breaking lines (writing tree to treefile) */
int Info_optn;        /* show more info about computational status */
int Maxbrnch;         /* 2*Maxspc - 3 */
int minedge;          /* edge with minimum edgeinfo */
int nextedge;         /* number of edges in the current tree */
int nextleaf;         /* next leaf to add to tree */
int outgroup;         /* outgroup */
int splitlength;      /* length of one entry in splitpatterns */
int Tree_optn;        /* generate treefile */
ivector trueID;       /* leaf -> taxon on this leaf */
ONEEDGE *edge;        /* vector with all the edges of the tree */
time_t Starttime;     /* start time */
time_t Stoptime;      /* stop time */
time_t time1;         /* other timer I */
time_t time2;         /* other timer II */
ulint *splitcomp;     /* bipartition storage */
ulint *splitfreqs;    /* frequencies of all different splits of all trees */
ulint *splitpatterns; /* all different splits of all trees */
ulint *splitsizes;    /* size of all different splits of all trees */
ulint badqs;          /* number of bad quartets */
ulint consincluded;   /* number of included biparts in the consensus tree */
ulint Currtrial;      /* counter for quartet puzzling trees */
ulint maxbiparts;     /* space is reserved for that many bipartitions */
ulint mininfo;        /* value of edgeinfo on minedge */
ulint numbiparts;     /* number of different bipartitions */
ulint Numquartets;    /* number of quartets */
ulint Numtrial;       /* number of quartet puzzling trees */
unsigned char *quartetinfo; /* place where quartets are stored */

/* prototypes */
int ulintcmp();         /* compare function for quicksort */
int fputid(FILE *ofp, int t);


/******************************************************************************/
/* random numbers generator  (Numerical recipes)                              */
/******************************************************************************/

/* definitions */
#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

/* variable */
long idum;

double randomunitintervall()
/* Long period (> 2e18) random number generator. Returns a uniform random
   deviate between 0.0 and 1.0 (exclusive of endpoint values).

   Source:
   Press et al., "Numerical recipes in C", Cambridge University Press, 1992
   (chapter 7 "Random numbers", ran2 random number generator) */
{
	int j;
	long k;
	static long idum2=123456789;
	static long iy=0;
	static long iv[NTAB];
	double temp;

	if (idum <= 0) {
		if (-(idum) < 1)
			idum=1;
		else
			idum=-(idum);
		idum2=(idum);
		for (j=NTAB+7;j>=0;j--) {
			k=(idum)/IQ1;
			idum=IA1*(idum-k*IQ1)-k*IR1;
			if (idum < 0)
				idum += IM1;
			if (j < NTAB)
				iv[j] = idum;
		}
		iy=iv[0];
	}
	k=(idum)/IQ1;
	idum=IA1*(idum-k*IQ1)-k*IR1;
	if (idum < 0)
		idum += IM1;
	k=idum2/IQ2;
	idum2=IA2*(idum2-k*IQ2)-k*IR2;
	if (idum2 < 0)
		idum2 += IM2;
	j=iy/NDIV;
	iy=iv[j]-idum2;
	iv[j] = idum;
	if (iy < 1)
		iy += IMM1;
	if ((temp=AM*iy) > RNMX)
		return RNMX;
	else
		return temp;
}

void initrandom()
{
	srand((unsigned) time(NULL));
	idum=-(long) rand();
}

/* returns a random integer in the range [0; n - 1] */
int randominteger(int n)
{
	return (int) floor(randomunitintervall()*n);
}


/******************************************************************************/
/* functions for representing and building trees                              */
/******************************************************************************/

/* initialize tree with the following starting configuration

                 2
          0  +------- C(=2)
  A(=0) -----+
             +------- B(=1)
                 1
                                */
void inittree()
{
	int i;

	/* allocate the memory for the whole tree */
	
	/* allocate memory for vector with all the edges of the tree */
	edge = (ONEEDGE *) calloc(Maxbrnch, sizeof(ONEEDGE) );
	if (edge == NULL) maerror("edge [inittree]");

	/* allocate memory for vector with edge numbers of leaves */ 
	edgeofleaf = (int *) calloc(Maxspc, sizeof(int) );
	if (edgeofleaf == NULL) maerror("edgeofleaf [inittree]");
	
	/* allocate memory for all the edges the edge map */
	for (i = 0; i < Maxbrnch; i++) {
		edge[i].edgemap = (int *) calloc(Maxbrnch, sizeof(int) );
		if (edge[i].edgemap == NULL) maerror("edgemap [inittree]");
	}
		
	/* number all edges */
	for (i = 0; i < Maxbrnch; i++) edge[i].numedge = i;
	
	/* initialize tree */
	
	nextedge = 3;
	nextleaf = 3;
	
	/* edge maps */
	(edge[0].edgemap)[0] = 0; /* you are on the right edge */
	(edge[0].edgemap)[1] = 4; /* go down left for leaf 1 */
	(edge[0].edgemap)[2] = 5; /* go down right for leaf 2 */
	(edge[1].edgemap)[0] = 1; /* go up for leaf 0 */
	(edge[1].edgemap)[1] = 0; /* you are on the right edge */
	(edge[1].edgemap)[2] = 3; /* go up/down right for leaf 2 */
	(edge[2].edgemap)[0] = 1; /* go up for leaf 0 */
	(edge[2].edgemap)[1] = 2; /* go up/down left for leaf 1 */
	(edge[2].edgemap)[2] = 0; /* you are on the right edge */
	
	/* interconnection */
	edge[0].up = NULL;
	edge[0].downleft = &edge[1];
	edge[0].downright = &edge[2];
	edge[1].up = &edge[0];
	edge[1].downleft = NULL;
	edge[1].downright = NULL;
	edge[2].up = &edge[0];
	edge[2].downleft = NULL;
	edge[2].downright = NULL;
	
	/* edges of leaves leaves */
	edgeofleaf[0] = 0;
	edgeofleaf[1] = 1;
	edgeofleaf[2] = 2;
}

/* add next leaf on the specified edge */
void addnextleaf(int dockedge)
{	
	int i;

	if (dockedge >= nextedge) {
		fprintf(stderr, "\nERROR: Trying to add leaf %d", nextleaf);
		fprintf(stderr, " to nonexisting edge %d [addnextleaf]!\n", dockedge);
		exit(1);
	}
	
	if (nextleaf >= Maxspc) {
		fprintf(stderr, "\nERROR: Trying to add leaf %d", nextleaf);
		fprintf(stderr, " to a tree with %d leaves [addnextleaf]!\n", Maxspc);
		exit(1);
	}
		
	/* necessary change in edgeofleaf if dockedge == edgeofleaf[0] */
	if (edgeofleaf[0] == dockedge) edgeofleaf[0] = nextedge;
	
	/* adding nextedge to the tree */
	edge[nextedge].up = edge[dockedge].up;
	edge[nextedge].downleft = &edge[dockedge];
	edge[nextedge].downright = &edge[nextedge+1];
	edge[dockedge].up = &edge[nextedge];

	if (edge[nextedge].up != NULL) {
		if ( ((edge[nextedge].up)->downleft) == &edge[dockedge] ) 
			(edge[nextedge].up)->downleft  = &edge[nextedge];
		else   
			(edge[nextedge].up)->downright = &edge[nextedge];
	}

	/* adding nextedge + 1 to the tree */
	edge[nextedge+1].up = &edge[nextedge];
	edge[nextedge+1].downleft = NULL;
	edge[nextedge+1].downright = NULL;
	edgeofleaf[nextleaf] = nextedge+1;
	
	/* the two new edges get info about the old edges */
	/* nextedge */
	for (i = 0; i < nextedge; i++) {
		switch ( (edge[dockedge].edgemap)[i] ) {
			
			/* down right changes to down left */
			case 5:		(edge[nextedge].edgemap)[i] = 4;
						break;
						
			/* null changes to down left */
			case 0:		(edge[nextedge].edgemap)[i] = 4;
						break;
			
			default: 	(edge[nextedge].edgemap)[i] =
							(edge[dockedge].edgemap)[i];
						break;
		}
	}

	/* nextedge + 1 */
	for (i = 0; i < nextedge; i++) {
		switch ( (edge[dockedge].edgemap)[i] ) {
			
			/* up/down left changes to up */
			case 2:		(edge[nextedge+1].edgemap)[i] = 1;
						break;
			
			/* up/down right changes to up */		
			case 3:		(edge[nextedge+1].edgemap)[i] = 1;
						break;
						
			/* down left changes to up/down left */			
			case 4:		(edge[nextedge+1].edgemap)[i] = 2;
						break;

			/* down right changes to up/down left */	
			case 5:		(edge[nextedge+1].edgemap)[i] = 2;
						break;
			
			/* null changes to up/down left */
			case 0:		(edge[nextedge+1].edgemap)[i] = 2;
						break;
			
			/* up stays up */
			default: 	(edge[nextedge+1].edgemap)[i] =
							(edge[dockedge].edgemap)[i];
						break;
		}
	}

	/* dockedge */
	for (i = 0; i < nextedge; i++) {
		switch ( (edge[dockedge].edgemap)[i] ) {
			
			/* up/down right changes to up */
			case 3:		(edge[dockedge].edgemap)[i] = 1;
						break;
						
			/* up/down left changes to up */
			case 2:		(edge[dockedge].edgemap)[i] = 1;
						break;
			
			default: 	break;
		}
	}

	/* all edgemaps are updated for the two new edges */
	/* nextedge */
	(edge[nextedge].edgemap)[nextedge] = 0;
	(edge[nextedge].edgemap)[nextedge+1] = 5; /* down right */
	
	/* nextedge + 1 */
	(edge[nextedge+1].edgemap)[nextedge] = 1; /* up */
	(edge[nextedge+1].edgemap)[nextedge+1] = 0;
	
	/* all other edges */
	for (i = 0; i < nextedge; i++) {
		(edge[i].edgemap)[nextedge] = (edge[i].edgemap)[dockedge];
		(edge[i].edgemap)[nextedge+1] = (edge[i].edgemap)[dockedge];
	}
	
	/* an extra for dockedge */
	(edge[dockedge].edgemap)[nextedge] = 1; /* up */
	(edge[dockedge].edgemap)[nextedge+1] = 3; /* up/down right */

	nextleaf++;
	nextedge = nextedge + 2;
}

/* free memory (to be called after inittree) */
void freetree()
{
	int i;

	for (i = 0; i < 2 * Maxspc - 3; i++) free(edge[i].edgemap);
	free(edge);
	free(edgeofleaf);	
}

/* clear all edgeinfos */
void resetedgeinfo()
{
	int i;
	
	for (i = 0; i < nextedge; i++)
		edge[i].edgeinfo = 0;
}

/* increment all edgeinfo between leaf A and B */
void incrementedgeinfo(int A, int B)
{	
	int curredge, finaledge, nextstep;
	
	if (A == B) return;
	
	finaledge = edgeofleaf[B];
	
	curredge = edgeofleaf[A];
	edge[curredge].edgeinfo = edge[curredge].edgeinfo + 1;
	
	while (curredge != finaledge) {
		nextstep = (edge[curredge].edgemap)[finaledge];
		switch (nextstep) {

			/* up */
			case 1:	curredge = (edge[curredge].up)->numedge;
					break;
				
			/* up/down left */
			case 2:	curredge = ((edge[curredge].up)->downleft)->numedge;
					break;

			/* up/down right */
			case 3:	curredge = ((edge[curredge].up)->downright)->numedge;
					break;
				
			/* down left */
			case 4:	curredge = (edge[curredge].downleft)->numedge;
					break;
				
			/* down right */
			case 5:	curredge = (edge[curredge].downright)->numedge;
					break;
				
		}
		edge[curredge].edgeinfo = edge[curredge].edgeinfo + 1;
	}	
}

/* prototype so that writeOTU knows about itself */
void writeOTU(FILE *outfp, int ed);

/* write OTU (recursive function) writes OTU sitting on edge ed */
void writeOTU(FILE *outfp, int ed)
{	
	int i;

	/* test whether we are on a leaf */
	if (edge[ed].downright == NULL && edge[ed].downleft == NULL) {
		for (i = 1; i < nextleaf; i++) {
			if (edgeofleaf[i] == ed) { /* i is the leaf of ed */
				column += fputid(outfp, trueID[i]);
				return;
			}
		}
	}

	/* we are NOT on a leaf */
	fprintf(outfp, "(");
	column++;
	writeOTU(outfp, edge[ed].downleft->numedge);
	fprintf(outfp, ", ");
	column++;
	column++;
	if (column > 55) {
		column = 2;
		fprintf(outfp, "\n  ");
	}
	writeOTU(outfp, edge[ed].downright->numedge);	
	fprintf(outfp, ")");
	column++;
}

/* write tree to treefile */
void writetree(FILE *outfp)
{	
	column = 1;
	fprintf(outfp, "(");
	column += fputid(outfp, trueID[0]) + 3;
	fprintf(outfp, ", ");
	writeOTU(outfp, edge[edgeofleaf[0]].downleft->numedge);
	column++;
	column++;
	fprintf(outfp, ", ");
	writeOTU(outfp, edge[edgeofleaf[0]].downright->numedge);	
	fprintf(outfp, ");\n");	
}

/* checks which edge has the lowest edgeinfo
   if there are several edges with the same lowest edgeinfo,
   one of them will be selected randomly */
void minimumedgeinfo()
{
	int i, k, howmany, randomnum;

	howmany = 1;
	minedge = 0;
	mininfo = edge[0].edgeinfo;
	for (i = 1; i < nextedge; i++)
		if (edge[i].edgeinfo <= mininfo) {
			if (edge[i].edgeinfo == mininfo) {
				howmany++;
			} else {
				minedge = i;
				mininfo = edge[i].edgeinfo;
				howmany = 1;
			}
		}
	
	if (howmany > 1) { /* draw random edge */
		randomnum = randominteger(howmany) + 1; /* 1 to howmany */
		i = -1;
		for (k = 0; k < randomnum; k++) {
			do {
				i++;
			} while (edge[i].edgeinfo != mininfo);
			minedge = i;
		}
	}
}


/******************************************************************************/
/* functions for computing the consensus tree                                 */
/******************************************************************************/

/* prototype needed for recursive function */
void makepart(int i, int curribrnch);

/* recursive function to get bipartitions */
void makepart(int i, int curribrnch)
{	
	int j;
	
	if ( edge[i].downright == NULL ||
		  edge[i].downleft == NULL) { /* if i is leaf */
			
			/* check out what leaf j sits on this edge i */		
			for (j = 1; j < Maxspc; j++) {
				if (edgeofleaf[j] == i) {
					biparts[curribrnch][trueID[j]] = '*';	
					return;
				}	
			}
	} else { /* still on inner branch */
		makepart(edge[i].downleft->numedge, curribrnch);
		makepart(edge[i].downright->numedge, curribrnch);
	}
}

/* compute bipartitions of current quartet puzzling tree */
void computebiparts()
{
	int i, j, curribrnch;
	
	curribrnch = -1;
	
	for (i = 0; i < Maxspc - 3; i++)
		for (j = 0; j < Maxspc; j++)
			biparts[i][j] = '.';

	for (i = 0; i < Maxbrnch; i++) {
		if (!(     edgeofleaf[0] == i    ||
		       edge[i].downright == NULL ||
		        edge[i].downleft == NULL) ) { /* check all inner branches */
			curribrnch++;
			makepart(i, curribrnch);
			
			/* make sure that the root is always a '*' */
			if (biparts[curribrnch][outgroup] == '.') {
				for (j = 0; j < Maxspc; j++) {
					if (biparts[curribrnch][j] == '.')
						biparts[curribrnch][j] = '*';
					else
						biparts[curribrnch][j] = '.';
				}
			}
		}
	}
}

/* print out the bipartition n of all different splitpatterns */
void printsplit(FILE *fp, ulint n)
{
	int i, j, col;
	ulint z;
	
	col = 0;
	for (i = 0; i < splitlength; i++) {
		z = splitpatterns[n*splitlength + i];
		for (j = 0; j < 32 && col < Maxspc; j++) {
			if (col % 10 == 0 && col != 0) fprintf(fp, " ");
			if (z & 1) fprintf(fp, ".");
			else fprintf(fp, "*");
			z = (z >> 1);
			col++;
		}
	}
}

/* make new entries for new different bipartitions and count frequencies */
void makenewsplitentries()
{
	int i, j, bpc, identical, idflag;
	ulint nextentry, obpc, bpsize;

	/* where the next entry would be in splitpatterns */
	nextentry = numbiparts;
	
	for (bpc = 0; bpc < Maxspc - 3; bpc++) { /* for every new bipartition */	
		/* convert bipartition into a more compact format */
		bpsize = 0;
		for (i = 0; i < splitlength; i++) {
			splitcomp[i] = 0;	
			for (j = 0; j < 32; j++) {
				splitcomp[i] = splitcomp[i] >> 1;
				if (i*32 + j < Maxspc)
					if (biparts[bpc][i*32 + j] == '.') {
						/* set highest bit */
						splitcomp[i] = (splitcomp[i] | 2147483648);
						bpsize++; /* count the '.' */
					}
			}
		}		
		/* compare to the *old* patterns */
		identical = FALSE;
		for (obpc = 0; (obpc < numbiparts) && (!identical); obpc++) {
			/* compare first partition size */
			if (splitsizes[obpc] == bpsize) idflag = TRUE;
			else idflag = FALSE;
			/* if size is identical compare whole partition */
			for (i = 0; (i < splitlength) && idflag; i++)
				if (splitcomp[i] != splitpatterns[obpc*splitlength + i])
					idflag = FALSE;
			if (idflag) identical = TRUE;
		}
		if (identical) { /* if identical increase frequency */
			splitfreqs[2*(obpc-1)]++;
		} else { /* create new entry */
			if (nextentry == maxbiparts) { /* reserve more memory */
				maxbiparts = maxbiparts + 2*Maxspc;
				splitfreqs = (ulint *) realloc(splitfreqs,
					2*maxbiparts * sizeof(ulint) );
				/* 2x: splitfreqs contains also an index (sorting!) */
				if (splitfreqs == NULL) maerror("splitfreqs");
				splitpatterns = (ulint *) realloc(splitpatterns,
					splitlength*maxbiparts * sizeof(ulint) );
				if (splitpatterns == NULL) maerror("splitpatterns");
				splitsizes = (ulint *) realloc(splitsizes,
					maxbiparts * sizeof(ulint) );
				if (splitsizes == NULL) maerror("splitsizes");
			}
			splitfreqs[2*nextentry] = 1; /* frequency */
			splitfreqs[2*nextentry+1] = nextentry; /* index for sorting */
			for (i = 0; i < splitlength; i++)
				splitpatterns[nextentry*splitlength + i] = splitcomp[i];
			splitsizes[nextentry] = bpsize;
			nextentry++;
		}
	}
	numbiparts = nextentry;
}


/******************************************************************************/
/* storing and evaluating quartet branching information                       */
/******************************************************************************/

/* general remarks:

	for a quartet with the taxa a, b, c, d there are
	three possible binary trees:
	
		1)  (a,b)-(c,d)
		2)  (a,c)-(b,d)
		3)  (a,d)-(b,c)
	
	For every quartet information about its branching structure is
	stored. With the functions  readquartet  and  writequartet
	this information can be accessed. For every quartet (a,b,c,d)
	with a < b < c < d (taxa) the branching information is encoded
	using 4 bits:
	
	value          8             4             2             1
	        +-------------+-------------+-------------+-------------+
	        |  not used   |   tree 3    |   tree 2    |   tree 1    |
	        +-------------+-------------+-------------+-------------+

	If the branching structure of the taxa correspond to one of the
	three trees the corresponding bit is set. If the branching structure
	is unclear because two of the three trees have the same maximum
	likelihood value the corresponding two bits are set. If the branching
	structure is completely unknown, all the bits are set (the highest
	bit is always cleared because it is not used).

*/

/* allocate memory for quartets */
void mallocquartets(int taxa)
{
	ulint nc, numch;

	/* number of quartets */
	Numquartets = (ulint) taxa*(taxa-1)*(taxa-2)*(taxa-3)/24;
	if (Numquartets % 2 == 0) { /* even number */
		numch = Numquartets/2;
	} else { /* odd number */
		numch = (Numquartets + 1)/2;
	}
	/* allocate memory */
	quartetinfo = (unsigned char *) malloc(numch * sizeof(unsigned char) );
	if (quartetinfo == NULL) maerror("mallocquartets");
	for (nc = 0; nc < numch; nc++) quartetinfo[nc] = 0;
}

/* free quartet memory */
void freequartets()
{	
	free(quartetinfo);
}

/* read quartet info - a < b < c < d */
unsigned char readquartet(int a, int b, int c, int d)
{
	ulint qnum;

	qnum = (ulint) a
			+ (ulint) b*(b-1)/2
			+ (ulint) c*(c-1)*(c-2)/6
			+ (ulint) d*(d-1)*(d-2)*(d-3)/24;
	if (qnum % 2 == 0) { /* even number */
		/* bits 0 to 3 */
		return (quartetinfo[qnum/2] & (unsigned char) 15);
	} else { /* odd number */
		/* bits 4 to 7 */
		return ((quartetinfo[(qnum-1)/2] & (unsigned char) 240)>>4);
	}
}

/* write quartet info - a < b < c < d, 0 <= info <= 15 */
void writequartet(int a, int b, int c, int d, unsigned char info)
{
	ulint qnum;

	qnum = (ulint) a
			+ (ulint) b*(b-1)/2
			+ (ulint) c*(c-1)*(c-2)/6
			+ (ulint) d*(d-1)*(d-2)*(d-3)/24;
	if (qnum % 2 == 0) { /* even number */
		/* bits 0 to 3 */
		quartetinfo[qnum/2] =
			((quartetinfo[qnum/2] & (unsigned char) 240) |
			(info & (unsigned char) 15));
	} else { /* odd number */
		/* bits 4 to 7 */
		quartetinfo[(qnum-1)/2] =
			((quartetinfo[(qnum-1)/2] & (unsigned char) 15) |
			((info & (unsigned char) 15)<<4));
	}
}

/* checks out all possible quartets */
void computeallquartets()
{	
	int a, b, c, i, maxml, badquartet;
	double d1, d2, d3, diff12, diff13, diff23, qc1, qc2;
	unsigned char qpbranching;
	ulint nq;

	nq = 0;
	badqs = 0;
	
	/* start timer - percentage of completed quartets */
	time(&time1);
	qc1 = 0.0;
	
	for (i = 3; i < Maxspc; i++) 
			for (a = 0; a < i - 2; a++)
					for (b = a + 1; b < i - 1; b++)
						for (c = b + 1; c < i; c++) {
							nq++;

							if(Info_optn) {
								fprintf(stderr,
									"\n        Checking quartet # %lu", nq);
								fprintf(stderr, "  [");
								fputid(stderr, a);
								fprintf(stderr, "-");
								fputid(stderr, b);
								fprintf(stderr, "-");
								fputid(stderr, c);
								fprintf(stderr, "-");
								fputid(stderr, i);
								fprintf(stderr, "]");
							} else { /* generate message every 15 minutes */
								/* check timer */
								time(&time2);
								if ( (time2 - time1) > 900) {
									/* every 900 seconds */
									/* percentage of completed quartets */
									qc2 = 100.*nq/Numquartets;
									fprintf(stderr,
										"\n        %.2f %% of all quartets ",
											qc2);
									fprintf(stderr,
										"completed (remaining time: %.0f ",
											(100.0-qc2) *
												(time2-time1)/60.0/(qc2-qc1));
									fprintf(stderr, "minutes)\n");
									time1 = time2;
									qc1 = qc2;
								}
							}
							
							/* maximum likelihood values */
							if (Approx_optn) { /* approximation */
								d1 = quartet_alklhd(a,b,c,i); /* (a,b)-(c,i) */
								d2 = quartet_alklhd(a,c,b,i); /* (a,c)-(b,i) */
								d3 = quartet_alklhd(a,i,b,c); /* (a,i)-(b,c) */
							} else { /* precise values */
								d1 = quartet_lklhd(a,b,c,i); /* (a,b)-(c,i) */
								d2 = quartet_lklhd(a,c,b,i); /* (a,c)-(b,i) */
								d3 = quartet_lklhd(a,i,b,c); /* (a,i)-(b,c) */
							}

							diff12 = (d1-d2)*(d1-d2);
							diff13 = (d1-d3)*(d1-d3);
							diff23 = (d2-d3)*(d2-d3);

							badquartet = FALSE;
							qpbranching = 0;

							/* looking for max(d1, d2, d3) */
							
							if (d1 < d2) { /* d2 > d1 */
								if (d2 < d3) { /* d3 > d2 > d1 */
									/* d3 maximum */ 
									writequartet(a, b, c, i, 4);
									maxml = 3;
								} else { /* d2 >= d3 > d1 */
									if (diff23 == 0.0) {
										/* d2 = d3 maximum */
										writequartet(a, b, c, i, 2 + 4);
										badquartet = TRUE;
									} else
										/* d2 maximum */
										writequartet(a, b, c, i, 2);
									maxml = 2;
								}
							} else { /* d1 >= d2 */
								if (d1 < d3) { /* d3 > d1 >= d2 */
									/* d3 maximum */
									writequartet(a, b, c, i, 4);
									maxml = 3;
								} else { /* d1 >= d2 && d1 >= d3 */
									qpbranching = 1; 
									if (diff12 == 0.0) {
										qpbranching = qpbranching + 2;
										badquartet = TRUE;
									}
									if (diff13 == 0.0) {
										qpbranching = qpbranching + 4;
										badquartet = TRUE;
									}
									writequartet(a, b, c, i, qpbranching);
									maxml = 1;
								}
							}

							/* check whether this quartet a good quartet */

							if (diff12 < diff13) {
								if (diff12 < diff23) {
									/* diff12 < diff23 && diff12 < diff13 */
									/* diff12 minimum */
									/* maxml should be 3 */
									if (maxml != 3) badquartet = TRUE;
								} else {
									/* diff23 =< diff12 < diff13 */	
									/* diff23 min */
									/* maxml should be 1 */
									if (maxml != 1) badquartet = TRUE;
								}
							} else {
								if (diff13 < diff23) { /* diff13 min*/
									/* maxml should be 2 */
									if (maxml != 2) badquartet = TRUE;
								} else { /* diff23 min */
									/* maxml should be 1 */
									if (maxml != 1) badquartet = TRUE;
								}
							}
		
							if (badquartet) {
								badqs++;
								if (Info_optn) fprintf(stderr,
									"    Bad quartet!");
							}
		
						}
	if(Info_optn) fprintf(stderr, "\n");
	fprintf(stderr,
		"\nINFO:  %lu quartets analysed - %lu bad quartets found (= %.1f %%)\n",
		nq, badqs, (double) badqs / (double) nq * 100);
}

/* check the branching structure between the leaves (not the taxa!)
   A, B, C, and I (A, B, C, I don't need to be ordered). As a result,
   the two leaves that are closer related to each other than to leaf I
   are found in chooseA and chooseB. If the branching structure is
   not uniquely defined, ChooseA and ChooseB are chosen randomly
   from the possible taxa */
void checkquartet(int A, int B, int C, int I)
{
	int i, j, a, b, taxon[5], leaf[5], ipos;
	unsigned char qresult;
	int notunique;

	/* The relationship between leaves and taxa is defined by trueID */
	taxon[1] = trueID[A]; /* taxon number */
	leaf[1] = A;          /* leaf number  */
	taxon[2] = trueID[B];
	leaf[2] = B;
	taxon[3] = trueID[C];
	leaf[3] = C;
	taxon[4] = trueID[I];
	leaf[4] = I;

	/* sort after taxa */
	/* Source: Numerical Recipes (PIKSR2.C) */
	for (j = 2; j <= 4; j++) {
		a = taxon[j];
		b = leaf[j];
		i = j-1;
		while (i > 0 && taxon[i] > a) {
			taxon[i+1] = taxon[i];
			leaf[i+1] = leaf[i];
			i--;
		}
		taxon[i+1] = a;
		leaf[i+1] = b;
	}

	/* where is leaf I ? */
	ipos = 1;
	while (leaf[ipos] != I) ipos++;

	/* look at taxa quartet */
	qresult = readquartet(taxon[1], taxon[2], taxon[3], taxon[4]);

	/* chooseA and chooseB */
	do {	
		switch (qresult) {
		
			/* one single branching structure */
		
			/* 001 */
			case 1:		if (ipos == 1 || ipos == 2) {
							chooseA = leaf[3];
							chooseB = leaf[4];
						} else {
							chooseA = leaf[1];
							chooseB = leaf[2];
						}
						notunique = FALSE;
						break;

			/* 010 */
			case 2:		if (ipos == 1 || ipos == 3) {
							chooseA = leaf[2];
							chooseB = leaf[4];
						} else {
							chooseA = leaf[1];
							chooseB = leaf[3];
						}
						notunique = FALSE;
						break;

			/* 100 */
			case 4:		if (ipos == 1 || ipos == 4) {
							chooseA = leaf[2];
							chooseB = leaf[3];
						} else {
							chooseA = leaf[1];
							chooseB = leaf[4];
						}
						notunique = FALSE;
						break;

			/* two possible branching structures */		

			/* 011 */
			case 3:		if (randominteger(2)) qresult = 1;
						else qresult = 2;
						notunique = TRUE;
						break;

			/* 101 */
			case 5:		if (randominteger(2)) qresult = 1;
						else qresult = 4;
						notunique = TRUE;
						break;

			/* 110 */
			case 6:		if (randominteger(2)) qresult = 2;
						else qresult = 4;
						notunique = TRUE;
						break;

			/* three possible branching structures */

			/* 111 */
			case 7:		qresult = (1 << randominteger(3)); /* 1, 2, or 4 */
						notunique = TRUE;
						break;

			default:	fprintf(stderr,
							"\nERROR: Program error [checkquartet]!\n");
						exit(1);
		}
	} while (notunique);

	return;
}


/******************************************************************************/
/* input and output handling                                                  */
/******************************************************************************/

/* read ten characters of current line as identifier */
void readid(FILE *infp, int t)
{
	int i, flag;

	for (i = 0; i < 10; i++) {
		Identif[t][i] = fgetc(infp);
		if (ferror(infp) || !isprint(Identif[t][i]) ) {
			fprintf(stderr, "\nERROR: I can't read the identifier");
			fprintf(stderr, " (taxon # %d)!\n", t + 1);
			exit(1);
		}
	}	
	/* convert leading blanks in taxon name to underscores */
	flag = FALSE;
	for (i = 9; i > -1; i--) {
		if (flag == FALSE) {
			if (Identif[t][i] != ' ') flag = TRUE; 
		} else {
			if (Identif[t][i] == ' ') Identif[t][i] = '_';
		}
	}
}

/* read next allowed character */
char readnextcharacter(FILE *ifp, int notu, int nsite)
{
	char c;

	do { /* ignore blanks and control characters except newline */
		if (fscanf(ifp, "%c", &c) != 1) {
			fprintf(stderr, "\nERROR: Missing character");
			fprintf(stderr, " (taxon # %d [", notu + 1);
			fputid(stderr, notu);
			fprintf(stderr, "], ");
			fprintf(stderr, "character # %d)!\n", nsite + 1);
			exit(1);
		}
	} while (c == ' ' || (iscntrl(c) && c != '\n'));
	if (c == '\n') return c;
	if (isacid(c) == FALSE) {
		fprintf(stderr, "\nERROR: Invalid character");
		if (isprint(c)) fprintf(stderr, " '%c'", c);
		fprintf(stderr, " (taxon # %d [", notu + 1);
		fputid(stderr, notu);
		fprintf(stderr, "], character # %d)!\n", nsite + 1);
		exit(1);
	}
	return c;
}

/* skip rest of the line */
void skiprestofline(FILE* ifp, int notu, int nsite)
{
	char c;

	/* read chars until the first newline */
	do{
		c = fgetc(ifp);
		if (ferror(ifp)) {
			fprintf(stderr, "\nERROR: Missing new line");
			fprintf(stderr, " (taxon # %d [", notu + 1);
			fputid(stderr, notu);
			fprintf(stderr, "], character # %d)!\n", nsite + 1);
			exit(1);
		}
	} while (c != '\n');
}

/* skip control characters and blanks */
void skipcntrl(FILE *ifp, int notu, int nsite)
{
	char c;
	fpos_t cpos;

	/* read over all control characters and blanks */
	do {
		/* current file position */
		if (fgetpos(ifp, &cpos) ) {
			fprintf(stderr, "\nERROR: File position not available");
			fprintf(stderr, " (taxon # %d [", notu + 1);
			fputid(stderr, notu);
			fprintf(stderr, "], character # %d)!\n", nsite + 1);
			exit(1);
		}
		c = fgetc(ifp);
		if (ferror(ifp)) {
			fprintf(stderr, "\nERROR: I can't read character ");
			fprintf(stderr, " (taxon # %d [", notu + 1);
			fputid(stderr, notu);
			fprintf(stderr, "], character # %d)!\n", nsite + 1);
			exit(1);
		}
	} while (iscntrl(c) || c == ' ');
	/* go one step back */
	if (fsetpos(ifp, &cpos) ) {
		fprintf(stderr, "\nERROR: File positioning error");
		fprintf(stderr, " (taxon # %d [", notu + 1);
		fputid(stderr, notu);
		fprintf(stderr, "], character # %d)!\n", nsite + 1);
		exit(1);
	}
}

/* read sequences of one data set */
void getseqs(FILE *ifp)
{
	int notu, sn, co, nsite, endofline, linelength, i;
	char c;
	cvector code;
	cmatrix seqchars;
	
	code = new_cvector(codelength);
	seqchars = new_cmatrix(Maxspc, Maxseqc);
	/* read all characters */
	nsite = 0; /* next site to be read */
	while (nsite < Maxseqc) {
		/* read first taxon */
		notu = 0;
		/* go to next true line */
		skiprestofline(ifp, notu, nsite);
		skipcntrl(ifp, notu, nsite);
		if (nsite == 0) readid(ifp, notu);
		endofline = FALSE;
		linelength = 0;		
		do {
			c = readnextcharacter(ifp, notu, nsite + linelength);
			if (c == '\n') endofline = TRUE;
			else if (nsite + linelength < Maxseqc) {
				seqchars[notu][nsite + linelength] = c;
				linelength++;
			} else {
				endofline = TRUE;
				skiprestofline(ifp, notu, nsite + linelength);
			}
		} while (!endofline);	
		if (linelength == 0) {
			fprintf(stderr, "\nERROR: Line with length zero ");
			fprintf(stderr, " (taxon # %d [", notu + 1);
			fputid(stderr, notu);
			fprintf(stderr, "], character # %d)!\n", nsite + 1);
			exit(1);
		}
		/* read other taxa */
		for (notu = 1; notu < Maxspc; notu++) {
			/* go to next true line */
			if (notu != 1) skiprestofline(ifp, notu, nsite);
			skipcntrl(ifp, notu, nsite);
			if (nsite == 0) readid(ifp, notu);
			for (i = nsite; i < nsite + linelength; i++) {
				c = readnextcharacter(ifp, notu, i);
				if (c == '\n') { /* too short */
					fprintf(stderr, "\nERROR: Line too short");
					fprintf(stderr, " (taxon # %d [", notu + 1);
					fputid(stderr, notu);
					fprintf(stderr, "], character # %d)!\n", i + 1);
					exit(1);
				} else {
					seqchars[notu][i] = c;
				}
			}
		}
		nsite = nsite + linelength;
	}
	/* decode characters */
	for (notu = 0; notu < Maxspc; notu++) {
		for (sn = 0; sn < Maxsite; sn++) {
			for (co = 0; co < codelength; co++) {
				c = seqchars[notu][sn*codelength + co];
				code[co] = c;
			}
			Seqchar[notu][sn] = code2int(code);
		}
	}
	free_cvector(code);
	free_cmatrix(seqchars);
}

/* initialize identifer array */
void initid(int t)
{
	int i, j;

	Identif = new_cmatrix(t, 10);
	for (i = 0; i < t; i++)
		for (j = 0; j < 10; j++)
			Identif[i][j] = ' ';
}

/* read first line of sequence data set */
void getsizesites(FILE *ifp)
{
	if (fscanf(ifp, "%d", &Maxspc) != 1) {
			fprintf(stderr, "\nERROR: I can't read the number of taxa!\n");
			exit(1);
	}
	if (fscanf(ifp, "%d", &Maxsite) != 1) {
			fprintf(stderr, "\nERROR: I can't read the number of sites!\n");
			exit(1);
	}
	Maxseqc = Maxsite; /* number of sequence characters per taxon */
	if (Maxsite % codelength != 0) {
			fprintf(stderr,
				"\nERROR: The number of sites must be a multiple of %d!\n",
					codelength);
			exit(1);
	} else {
		Maxsite = Maxsite / codelength; /* number of ML sites per taxon */
	}
	if (Maxspc < 4) {
		fprintf(stderr,
			"\nERROR: First line of input file can't specify only %d taxa!\n",
				Maxspc);
		exit(1);
	}
	if (Maxspc > 568) {
		fprintf(stderr,
			"\nERROR: Puzzle can't build quartet puzzling trees with more");
		fprintf(stderr, " than 568 taxa!\n");
		exit(1);
	}
	if (Maxsite < 1) {
		fprintf(stderr,
			"\nERROR: First line of input file can't specify zero sites!\n");
		exit(1);
	}
}

/* read one data set - PHYLIP interleaved */
void getdataset(FILE *ifp)
{
	initid(Maxspc);
	Seqchar = new_cmatrix(Maxspc, Maxsite);
	getseqs(ifp);
}

/* print identifier of specified taxon in full 10 char length */
void fputid10(FILE *ofp, int t)
{	
	int i;

	for (i = 0; i < 10; i++) fputc(Identif[t][i], ofp);
}

/* print identifier of specified taxon up to first space */
int fputid(FILE *ofp, int t)
{	
	int i;
	
	i = 0;
	while (Identif[t][i] != ' ' && i < 10) {
		fputc(Identif[t][i], ofp);
		i++;
	}
	return i;
}


/* set options interactively */
void setoptions()
{	
	int ch, i;
	double sumfreq;

	/* defaults */
	AlphaBeta = 4.0;    /* ratio of Ts and Tv rates           */
	AlphaYR = 1.0;      /* ratio of Y and R transitions       */
	Approx_optn = TRUE; /* approximate likelihoods            */
	data_optn = 0;      /* nucleotides as sequence input data */
	Dayhf_optn = FALSE; /* Dayhoff model                      */
	Frequ_optn = TRUE;  /* use empirical frequencies          */
	Info_optn = FALSE;  /* show status of computation         */
	Jtt_optn = TRUE;    /* JTT model                          */
	mtrev_optn = FALSE; /* mtREV model                        */
	nuc_optn = TRUE;    /* 4x4 nucleotide models              */
	Numtrial = 1000;    /* number of quartet puzzling trees   */
	outgroup = 0;       /* use first taxon as outgroup        */
	SH_optn = FALSE;    /* SH 16x16 nucleotide model          */
	sym_optn = FALSE;   /* symmetrize frequencies             */ 
	Tree_optn = 2;      /* generate treefile                  */

 	do {	
		fprintf(stderr, "\n\nSettings for this run of Puzzle (version %s):\n\n",
			VERSION);
		fprintf(stderr, "d          Type of sequence input data?  %s\n",
			(data_optn ? "Proteins" : "Nucleotides"));
		fprintf(stderr, "m          Model of sequence evolution?  ");
		
		if (data_optn == 0) { /* nucleotides */
			if (nuc_optn) {
				if(AlphaBeta == 1.0 && AlphaYR == 1.0)
					fprintf(stderr, "F (Felsenstein 1981)\n");
				if(AlphaBeta != 1.0 && AlphaYR == 1.0)
					fprintf(stderr, "HKY (Hasegawa et al. 1985)\n");	
				if(AlphaYR != 1.0)
					fprintf(stderr, "TN (Tamura-Nei 1993)\n");
				if (AlphaYR != 1.0 || AlphaBeta != 1.0)
					fprintf(stderr,
						"t        Transition/transversion ratio?  %.2f\n",
							AlphaBeta / 2);	
				if (AlphaYR != 1.0)
					fprintf(stderr,
 						"r      Y-transition/R-transition ratio?  %.2f\n",
							AlphaYR);
			}
			if (SH_optn) {
					fprintf(stderr, "SH (Schoeniger-von Haeseler 1994)\n");
					fprintf(stderr,
 						"t        Transition/transversion ratio?  %.2f\n",
							AlphaBeta / 2);	
			}	
		}

		if (data_optn == 1) { /* proteins */
			if (Dayhf_optn) fprintf(stderr, "Dayhoff (Dayhoff et al. 1978)\n");	
			if (Jtt_optn) fprintf(stderr, "JTT (Jones et al. 1992)\n");
			if (mtrev_optn) fprintf(stderr, "mtREV (Adachi-Hasegawa 1995)\n");
		}

		fprintf(stderr, "f       Use empirical base frequencies?  %s\n",
			(Frequ_optn ? "Yes" : "No, use specified frequencies"));
		if (data_optn == 0 && SH_optn)
			fprintf(stderr, "s       Symmetrize doublet frequencies?  %s\n",
				(sym_optn ? "Yes" : "No"));
		fprintf(stderr, "a       Approximate maximum likelihood?  %s\n",
				(Approx_optn ? "Yes" : "No"));
		fprintf(stderr, "n     Number of quartet puzzling trees?  %lu\n",
				Numtrial);
		/*fprintf(stderr, "o                Use as outgroup taxon?  %d\n",
				outgroup + 1);*/
		/*fprintf(stderr, "w              Write trees to treefile?  ");
		if (Tree_optn == 0) fprintf(stderr, "No\n");
		if (Tree_optn == 1) fprintf(stderr, "Consensus tree\n");
		if (Tree_optn == 2) fprintf(stderr, "All trees\n");*/
		fprintf(stderr, "i          Show details of computation?  %s\n",
				(Info_optn ? "Yes" : "No"));
		fprintf(stderr, "\nStart with analysis? ");
		fprintf(stderr, "(type y or choose from the menu) ");
		
		/* read one char */
		ch = getchar();
		if (ch != '\n') {
			do ;
			while (getchar() != '\n');
		}
		ch = tolower(ch);
		
		switch (ch) {

			case '\n':	break;

			case 'a':	Approx_optn = 1 - Approx_optn;
						break;
			
			/*case 'w':	Tree_optn = 1 + Tree_optn;
						if (Tree_optn == 3) Tree_optn = 0;
						break;*/

			case 'd':	data_optn = 1 - data_optn;
						if (!Frequ_optn) {
							Frequ_optn = TRUE;
							free_dvector(Freqext);
						}
						break;
			
			case 'f':	if (Frequ_optn) {
							Frequ_optn = FALSE;
							Freqext = new_dvector(gettpmradix());
							
							fprintf(stderr,
								"\n\nPlease enter base frequencies (in %%): ");
							fprintf(stderr, "pi(%s) = ", int2code(0));
							scanf("%lf", &(Freqext[0]));
							do ;
							while (getchar() != '\n');
							Freqext[0] = Freqext[0]/100.0;
							if (Freqext[0] < 0.0) {
								fprintf(stderr, "\n\nERROR: Frequencies");
								fprintf(stderr, " can't be negative!\n");
								Frequ_optn = TRUE;
								free_dvector(Freqext);
								break;
							}
							sumfreq = Freqext[0];
							if (sumfreq > 1.0) {
								fprintf(stderr, "\n\nERROR: The sum of ");
								fprintf(stderr, "the frequencies can't");
								fprintf(stderr, " exceed 100 %%!\n");
								Frequ_optn = TRUE;
								free_dvector(Freqext);
								break;
							}
							for (i = 1; i < gettpmradix() - 1; i++) {
								fprintf(stderr, "                            ");
								fprintf(stderr,
									"          pi(%s) = ", int2code(i));
								scanf("%lf", &(Freqext[i]));
								do ;
								while (getchar() != '\n');
								Freqext[i] = Freqext[i]/100.0;
								if (Freqext[i] < 0.0) {
									fprintf(stderr, "\n\nERROR: Frequencies");
									fprintf(stderr, " can't be negative!\n");
									Frequ_optn = TRUE;
									free_dvector(Freqext);
									break;
								}
								sumfreq = sumfreq + Freqext[i];
								if (sumfreq > 1.0) {
									fprintf(stderr, "\n\nERROR: The sum of ");
									fprintf(stderr, "the frequencies can't");
									fprintf(stderr, " exceed 100 %%!\n");
									Frequ_optn = TRUE;
									free_dvector(Freqext);
									break;
								}
							}
							Freqext[gettpmradix()-1] = 1.0 - sumfreq;

						} else {
							Frequ_optn = TRUE;
							free_dvector(Freqext);
						}
						break;
			
			case 's':	if (data_optn == 0 && SH_optn) {
							sym_optn = 1 - sym_optn;
						} else {
							fprintf(stderr,
								"\n\nThis is not a possible option!\n");
						}
						break;
			
			case 'i':	Info_optn = 1 - Info_optn;
						break;

			case 'n':	fprintf(stderr,
							"\n\nNumber of quartet puzzling trees: ");
						scanf("%lu", &Numtrial);
						do ;
						while (getchar() != '\n');
						if (Numtrial < 1) {
							fprintf(stderr, "\n\nERROR: The number of trees");
							fprintf(stderr, " can't be smaller than one!\n");
							Numtrial = 100;
						}
						break;
						
			/*case 'o':	fprintf(stderr,
							"\n\nTaxon number of the outgroup: ");
						scanf("%d", &outgroup);
						do ;
						while (getchar() != '\n');
						if (outgroup < 1) {
							fprintf(stderr,
								"\n\nERROR: The taxon number of the out");
							fprintf(stderr,
								"group can't be smaller than one!\n");
							outgroup = 1;
						}
						outgroup = outgroup - 1;
						break;*/
	
			case 'm':	if (data_optn == 0) { /* nucleotide data */
							if(AlphaBeta == 1.0 && AlphaYR == 1.0 && nuc_optn) {
								/* F -> HKY */
								AlphaBeta = 4.0;
								AlphaYR = 1.0;
								nuc_optn = TRUE;
								SH_optn = FALSE;
								break;
							}
							if(AlphaBeta != 1.0 && AlphaYR == 1.0 && nuc_optn) {
								/* HKY -> TN */
								AlphaBeta = 4.0;
								AlphaYR = 2.0;
								nuc_optn = TRUE;
								SH_optn = FALSE;
								break;
							}
							if(AlphaYR != 1.0 && nuc_optn) {
								/* TN -> SH */
								AlphaBeta = 4.0;
								AlphaYR = 1.0;
								nuc_optn = FALSE;
								SH_optn = TRUE;
								if (!Frequ_optn) {
									Frequ_optn = TRUE;
									free_dvector(Freqext);
								}
								break;
							}
							if(SH_optn) {
								/* SH -> JC */
								AlphaBeta = 1.0;
								AlphaYR = 1.0;
								nuc_optn = TRUE;
								SH_optn = FALSE;
								if (!Frequ_optn) {
									Frequ_optn = TRUE;
									free_dvector(Freqext);
								}
								break;
							}
							break;
						}
						if (data_optn == 1) { /* protein data */
							if (Dayhf_optn) {
								/* Dayhoff -> JTT */
								Dayhf_optn = FALSE;
								Jtt_optn = TRUE;
								mtrev_optn = FALSE;
								break;
							}
							if (Jtt_optn) {
								/* JTT -> mtREV */
								Dayhf_optn = FALSE;
								Jtt_optn = FALSE;
								mtrev_optn = TRUE;
								break;
							}
							if (mtrev_optn) {
								/* mtREV -> Dayhoff */
								Dayhf_optn = TRUE;
								Jtt_optn = FALSE;
								mtrev_optn = FALSE;
								break;
							}
							break;
						}
						break;
						
			case 't':	if ((AlphaBeta == 1.0 && AlphaYR == 1.0 && nuc_optn) ||
							(AlphaBeta == 1.0 && SH_optn)) {
							fprintf(stderr,
								"\n\nThis is not a possible option!\n");
						} else {
							fprintf(stderr,
								"\n\nExpected transition/transversion ratio: ");
							scanf("%lf", &AlphaBeta);
							do ;
							while (getchar() != '\n');
							AlphaBeta = 2 * AlphaBeta;
							if (AlphaBeta < 0) {
								fprintf(stderr,
									"\n\nERROR: The transition/transversion");
								fprintf(stderr,
									" ratio can't be negative!\n");
								AlphaBeta = 4.0;
							}
							if (AlphaBeta == 0) {
								fprintf(stderr,
									"\n\nERROR: The transition/transversion");
								fprintf(stderr,
									" ratio can't be zero!\n");
								AlphaBeta = 4.0;
							}
						}
						break;

			case 'r':	if (AlphaYR == 1.0 || !nuc_optn) {
							fprintf(stderr,
								"\n\nThis is not a possible option!\n");
						} else {
							fprintf(stderr,
								"\n\nPyrimidine transition/Purine");
							fprintf(stderr, " transition ratio: ");
							scanf("%lf", &AlphaYR);
							do ;
							while (getchar() != '\n');
							if (AlphaYR < 0) {
								fprintf(stderr,
									"\n\nERROR: The Y-transition/R-transition");
								fprintf(stderr,
									" ratio can't be negative!\n");
								AlphaYR = 2.0;
							}
							if (AlphaYR == 0) {
								fprintf(stderr,
									"\n\nERROR: The Y-transition/R-transition");
								fprintf(stderr,
									" ratio can't be zero!\n");
								AlphaYR = 2.0;
							}
						}
						break;
						
			case 'y':	break;

			default:	fprintf(stderr, "\n\nThis is not a possible option!\n");
						break;
		}
	} while (ch != 'y');

	fprintf(stderr, "\n\n");
}

/* open files */
void openfiles()
{	
	char inputline[64];
	
	/* open sequencefile  */
    if ((seqfp = fopen(INFILE, "r")) == NULL) {
   		fprintf(stderr,
   			"\nERROR: I can't open the sequence input file '%s'!\n",
   				INFILE);
		fprintf(stderr,
			"       Please enter another name for this file: ");
		gets(inputline);
		if ((seqfp = fopen(inputline, "r")) == NULL) {
			fprintf(stderr, "\nERROR: I can't open this file either!\n");
			exit(1);
		}
	}

	if ((ofp = fopen(OUTFILE, "w")) == NULL) {
		fprintf(stderr,
			"\nERROR: I can't create the output file '%s'!\n", OUTFILE);
		fprintf(stderr,
			"       Please enter another name for this file: ");
		gets(inputline);
		if ((ofp = fopen(inputline, "w")) == NULL) {
			fprintf(stderr, "\nERROR: I can't create this file either!\n");
			exit(1);
		}
	}
	if (Tree_optn) {
		if ((tfp = fopen(TREEFILE, "w")) == NULL) {
			fprintf(stderr,
				"\nERROR: I can't create the treefile '%s'!\n", TREEFILE);
			fprintf(stderr,
				"       Please enter another name for this file: ");
			gets(inputline);
			if ((tfp = fopen(inputline, "w")) == NULL) {
				fprintf(stderr, "\nERROR: I can't create this file either!\n");
				exit(1);
			}
		}
	}
}

/* close files */
void closefiles()
{	
	fclose(seqfp);
	fclose(ofp);
	if (Tree_optn) fclose(tfp);
}

/* show ML base frequencies */
void showfrequencies()
{	
	int i;
	
	fprintf(stderr, "\nINFO:  ML base frequencies: pi(%s) = %.1f %%\n",
		int2code(0), Freqtpm[0]*100);
				
	for (i = 1; i < gettpmradix(); i++)
		fprintf(stderr, "                            pi(%s) = %.1f %%\n",
			int2code(i), Freqtpm[i]*100);
}

/* show Ts/Tv ratio */
void showtstvratio()
{
	double AlphaYBeta, AlphaRBeta, piR, piY, num, denom;
	
	if (nuc_optn == TRUE) { /* 4x4 nucs */
		piR = Freqtpm[0] + Freqtpm[2];
		piY = Freqtpm[1] + Freqtpm[3];
		AlphaRBeta = 2 * AlphaBeta / (1 + AlphaYR);
		AlphaYBeta = AlphaRBeta * AlphaYR;
		tstvratio = (AlphaRBeta*Freqtpm[0]*Freqtpm[2] +
					 AlphaYBeta*Freqtpm[1]*Freqtpm[3])/(piR * piY);
	} else { /* 16x16 nucs */
		num = Freqtpm[0]*Freqtpm[2] + Freqtpm[1]*Freqtpm[3] +
			Freqtpm[4]*Freqtpm[6] + Freqtpm[5]*Freqtpm[7] +
			Freqtpm[0]*Freqtpm[8] + Freqtpm[1]*Freqtpm[9] +
			Freqtpm[2]*Freqtpm[10] + Freqtpm[8]*Freqtpm[10] +
			Freqtpm[3]*Freqtpm[11] + Freqtpm[9]*Freqtpm[11] +
			Freqtpm[4]*Freqtpm[12] + Freqtpm[5]*Freqtpm[13] +
			Freqtpm[6]*Freqtpm[14] + Freqtpm[12]*Freqtpm[14] +
			Freqtpm[7]*Freqtpm[15] + Freqtpm[13]*Freqtpm[15];
		denom = Freqtpm[0]*Freqtpm[1] + Freqtpm[1]*Freqtpm[2] +
			Freqtpm[0]*Freqtpm[3] + Freqtpm[2]*Freqtpm[3] +
			Freqtpm[0]*Freqtpm[4] + Freqtpm[1]*Freqtpm[5] +
			Freqtpm[4]*Freqtpm[5] + Freqtpm[2]*Freqtpm[6] +
			Freqtpm[5]*Freqtpm[6] + Freqtpm[3]*Freqtpm[7] +
			Freqtpm[4]*Freqtpm[7] + Freqtpm[6]*Freqtpm[7] +
			Freqtpm[4]*Freqtpm[8] + Freqtpm[5]*Freqtpm[9] +
			Freqtpm[8]*Freqtpm[9] + Freqtpm[6]*Freqtpm[10] +
			Freqtpm[9]*Freqtpm[10] + Freqtpm[7]*Freqtpm[11] +
			Freqtpm[8]*Freqtpm[11] + Freqtpm[10]*Freqtpm[11] +
			Freqtpm[0]*Freqtpm[12] + Freqtpm[8]*Freqtpm[12] +
			Freqtpm[1]*Freqtpm[13] + Freqtpm[9]*Freqtpm[13] +
			Freqtpm[12]*Freqtpm[13] + Freqtpm[2]*Freqtpm[14] +
			Freqtpm[10]*Freqtpm[14] + Freqtpm[13]*Freqtpm[14] +
			Freqtpm[3]*Freqtpm[15] + Freqtpm[11]*Freqtpm[15] +
			Freqtpm[12]*Freqtpm[15] + Freqtpm[14]*Freqtpm[15];
		tstvratio = AlphaBeta * num/denom;
	}
	fprintf(stderr,
		"\nINFO:  Expected transition/transversion ratio: %.2f\n",
			tstvratio);
}

void writeoutputfile(FILE *ofp)
{
	int i;
	ulint li;

	fprintf(ofp, "QUARTET PUZZLING\n\n");
	fprintf(ofp, "Puzzle (version %s) session of %s",
		VERSION, asctime(localtime(&Starttime)) );
	fprintf(ofp, "Runtime: %lu seconds (= %.1f minutes = %.1f hours)\n\n\n",
		Stoptime - Starttime, (Stoptime - Starttime)/60.,
		(Stoptime - Starttime)/3600.);
	fprintf(ofp, "MAXIMUM LIKELIHOOD MODEL\n\n");
	fprintf(ofp, "Sequence input data: %d taxa with %d sites (%s)\n",
		Maxspc, Maxseqc, (data_optn ? "proteins" : "nucleotides"));
	fprintf(ofp, "Model of sequence evolution: ");
	if (data_optn == 0) { /* nucleotides */
		if (nuc_optn) {
			if(AlphaBeta == 1.0 && AlphaYR == 1.0)
				fprintf(ofp, "F (Felsenstein 1981)\n");
			if(AlphaBeta != 1.0 && AlphaYR == 1.0)
				fprintf(ofp, "HKY (Hasegawa et al. 1985)\n");	
			if(AlphaYR != 1.0)
				fprintf(ofp, "TN (Tamura-Nei 1993)\n");
			if (AlphaYR != 1.0 || AlphaBeta != 1.0)
				fprintf(ofp, "Transition/transversion ratio: %.2f\n",
					AlphaBeta / 2);	
			if (AlphaYR != 1.0)
				fprintf(ofp, "Y-transition/R-transition ratio: %.2f\n",
					AlphaYR);
		}
		if (SH_optn) {
			fprintf(ofp, "SH (Schoeniger-von Haeseler 1994)\n");
			fprintf(ofp, "Transition/transversion ratio: %.2f\n",
				AlphaBeta / 2);
		}	
	}
	if (data_optn == 1) { /* proteins */
		if (Dayhf_optn) fprintf(ofp, "Dayhoff (Dayhoff et al. 1978)\n");	
		if (Jtt_optn) fprintf(ofp, "JTT (Jones et al. 1992)\n");
		if (mtrev_optn) fprintf(ofp, "mtREV (Adachi-Hasegawa 1995)\n");
	}
	fprintf(ofp, "Base frequencies: pi(%s) = %.1f %%\n",
		int2code(0), Freqtpm[0]*100);		
	for (i = 1; i < gettpmradix(); i++)
		fprintf(ofp, "                  pi(%s) = %.1f %%\n",
			int2code(i), Freqtpm[i]*100);
	if (Frequ_optn) fprintf(ofp, "Base frequencies are empirical");
	else fprintf(ofp, "Base frequencies have been specified");
	if (data_optn == 0 && SH_optn && sym_optn)
		fprintf(ofp, " and symmetrized\n");
	else fprintf(ofp, "\n");
	if (data_optn == 0)
		fprintf(ofp, "Expected transition/transversion ratio: %.2f\n",
			tstvratio);
	fprintf(ofp, "\n\nQUARTET TREES\n\n");
	fprintf(ofp, "Analysed quartets: %lu\n", Numquartets);
	fprintf(ofp, "Bad quartets: %lu (= %.1f %%)\n",
		badqs, (double) badqs / (double) Numquartets * 100);
	fprintf(ofp, "Quartet trees are based on %s ",
		(Approx_optn ? "approximate" : "precise"));
	fprintf(ofp, "maximum likelihood values\n\n\n");
	fprintf(ofp, "MAJORITY RULE CONSENSUS TREE\n\n");
	fprintf(ofp, "Among the %lu generated quartet puzzling trees the following\n",
		Numtrial);
	fprintf(ofp, "bipartitions occured with a frequency of more than 50 %%.\n");
	fprintf(ofp, "These bipartitions define a majority rule consensus tree.\n");
	fprintf(ofp, "(If you want to view this tree graphically feed all the quartet\n");
	fprintf(ofp, "puzzling trees [qptrees] into the PHYLIP CONSENSE program ...)\n");
	fprintf(ofp, "\nBipartitions included in the consensus tree\n");
	fprintf(ofp, "(taxa in the order of the input file):\n\n");
	for (li = 0; li < consincluded; li++) {
		fprintf(ofp, " ");
		printsplit(ofp, splitfreqs[2*li+1]);
		fprintf(ofp, "  : %.2f %%\n",
			(double) splitfreqs[2*li]/Numtrial*100.0);
	}
	if (consincluded == 0) fprintf(ofp, " None (no bipartition included)\n");
	fprintf(ofp, "\nBipartitions not included in the consensus tree\n");
	fprintf(ofp, "(taxa in the order of the input file):\n\n");
	for (li = consincluded; li < numbiparts; li++) {
		fprintf(ofp, " ");
		printsplit(ofp, splitfreqs[2*li+1]);
		fprintf(ofp, "  : %.2f %%\n", (double) splitfreqs[2*li]/Numtrial*100.0);
	}
	if (consincluded == numbiparts)
		fprintf(ofp, " None (all bipartitions are included)\n");
	fprintf(ofp, "\nTaxa in the order of the input file:\n\n");
	for (i = 0; i < Maxspc; i++) {
		fprintf(ofp, " ");
		fputid(ofp, i);
		fprintf(ofp, "\n");
	}
}


/******************************************************************************/
/* main part                                                                  */
/******************************************************************************/

main()
{	
	int i, j, a, b, c, k, ID;
	ulint nqused;
	ivector freeID;
	double tc1, tc2;

	/* start */
	fprintf(stderr, "Reconstructing tree topologies with quartet puzzling\n");
	
	/* set options in interaction with the user */
	setoptions();

	/* initialize random numbers generator */
	initrandom();
	
	/* open files and get sequences */
	fprintf(stderr, "\nINFO:  Reading and processing sequence input data\n");
	openfiles();
	codelength = getcodelength(); /* doublets ? */
	getsizesites(seqfp);
	getdataset(seqfp);
	if (outgroup >= Maxspc) {
		fprintf(stderr,
			"\nERROR: Specified outgroup taxon doesn't exist!\n");
		fprintf(stderr, 
			"       I will use first taxon as outgroup instead!\n");
		outgroup = 0;
	}
	Maxbrnch = 2*Maxspc - 3;
	fprintf(stderr, "\nINFO:  Sequence input data: %d taxa with %d sites\n",
		Maxspc, Maxseqc);
	
	/* initialise ML */	
	qmlstart();

	/* show base frequencies */
	showfrequencies();
	
	/* show expected Ts/Tv ratio */
	if (data_optn == 0) showtstvratio();

	/* allocate variables used for randomizing input order */
	trueID = new_ivector(Maxspc);
	freeID = new_ivector(Maxspc);
	
	/* allocate memory for quartets */
	fprintf(stderr, "\nINFO:  Preparing for quartet information storage\n");
	mallocquartets(Maxspc);
	
	/* prepare for consensus tree analysis */
	biparts = new_cmatrix(Maxspc-3, Maxspc);
	if (Maxspc % 32 == 0)
		splitlength = Maxspc/32;
	else splitlength = (Maxspc + 32 - (Maxspc % 32))/32;
	numbiparts = 0; /* no pattern stored so far */
	maxbiparts = 0; /* no memory reserved so far */
	splitfreqs = NULL;
	splitpatterns = NULL;
	splitsizes = NULL;
	splitcomp = (ulint *) malloc(splitlength * sizeof(ulint) );
	if (splitcomp == NULL) maerror("splitcomp");
	
	/* start main timer */
	time(&Starttime);
		
	/* compute quartets */
	fprintf(stderr, "\nINFO:  Computing quartet maximum likelihood trees\n");
	computeallquartets();
	
	fprintf(stderr,
		"\nINFO:  Starting to build %lu quartet puzzling trees\n", Numtrial);
	
	/* start timer - percentage of completed trees */
	time(&time1);
	tc1 = 0.0;
	
	for (Currtrial = 0; Currtrial < Numtrial; Currtrial++) {
			
		/* randomize input order */		
		for (k = 0; k < Maxspc; k++) freeID[k] = TRUE;
		for (k = 0; k < Maxspc; k++) {
			/* random integer in the range [0, Maxspc-1-k] */
			ID = randominteger(Maxspc - k);
			i = -1;
			j = -1;
			do {
				i++;
				if (freeID[i] == TRUE) j++; 
			} while ( j != ID);
				trueID[k] = i;
				freeID[i] = FALSE;
		}

		/* initialize tree */
		inittree();
		if (Info_optn) {
			fprintf(stderr, "\n        Quartet puzzling tree # %lu:\n",
			Currtrial + 1);
			fprintf(stderr, "         leaf 0  > ");
			fputid10(stderr, trueID[0]);
			fprintf(stderr, " <  added  [0 quartets used]\n");
			fprintf(stderr, "         leaf 1  > ");
			fputid10(stderr, trueID[1]);
			fprintf(stderr, " <  added  [0 quartets used]\n");
			fprintf(stderr, "         leaf 2  > ");
			fputid10(stderr, trueID[2]);
			fprintf(stderr, " <  added  [0 quartets used]\n");
		}

		/* adding all other leafs */
		for (i = 3; i < Maxspc; i++) { 
					
			/* clear all edgeinfos */
			resetedgeinfo();
				
			/* clear counter of quartets */
			nqused = 0;

			/*
			 * core of quartet puzzling algorithm
			 */

			for (a = 0; a < nextleaf - 2; a++)
				for (b = a + 1; b < nextleaf - 1; b++)
					for (c = b + 1; c < nextleaf; c++) {

						/*	check which two *leaves* out of a, b, c
							are closer related to each other than
							to leaf i.
							These two leaves are chooseA and chooseB */

							checkquartet(a, b, c, i);

							incrementedgeinfo(chooseA, chooseB);
							nqused++;
							
							/* generate message every 15 minutes */
							if(!Info_optn) {
								/* check timer */
								time(&time2);
								if ( (time2 - time1) > 900) {
									/* every 900 seconds */
									/* percentage of completed trees */
									tc2 = 100.0*Currtrial/Numtrial + 
											100.0*nqused/Numquartets/Numtrial;
									fprintf(stderr,
										"\n        %2.2f %% of all trees ",
											tc2);
									fprintf(stderr,
										"completed (remaining time: %.0f",
											(100.0-tc2) *
												(time2-time1)/60.0/(tc2-tc1));
									fprintf(stderr, " minutes)\n");
									time1 = time2;
									tc1 = tc2;
								}
							}
					}

			/* find out which edge has the lowest edgeinfo */
			minimumedgeinfo();
				
			/* add the next leaf on minedge */
			addnextleaf(minedge);

			if(Info_optn) {
				fprintf(stderr, "         leaf %d  > ", i);
				fputid10(stderr, trueID[i]);
				fprintf(stderr, " <  added  [%lu quartets used]\n",
					nqused);
			}
		}
		
		/* writing trees */
		if(Tree_optn == 2) {
			writetree(tfp);
			/* flush tree buffer - useful if many qptrees are generated */
			if (Currtrial != 0 && Currtrial % 500 == 0) fflush(tfp);
		}
		
		/* compute bipartitions of current tree */
		computebiparts();
		makenewsplitentries();

		/* free memory before building the next tree */
		freetree();
	}

	/* sort bipartition frequencies */
	qsort(splitfreqs, numbiparts, 2*sizeof(ulint), ulintcmp);
	consincluded = 0;
	for (i = 0; i < numbiparts && i == consincluded; i++) {
		if (2*splitfreqs[2*i] > Numtrial) consincluded = i + 1;
	}

	/* stop timer */
	time(&Stoptime);
	
	/* write output file */
	writeoutputfile(ofp);
	
	/* close files */
	closefiles();
	if (Tree_optn == 1)
		fprintf(stderr, "\nINFO:  Consensus tree tree written to tree output file\n");
	if (Tree_optn == 2)
		fprintf(stderr,
			"\nINFO:  All %lu trees written to tree output file\n", Numtrial);

	/* runtime message */
	fprintf(stderr, "\nINFO:  This run took %lu seconds (= %.1f minutes)\n",
		Stoptime - Starttime, (Stoptime - Starttime)/60.);

	/* finish ML */
	qmlfinish();

	/* free memory */
	free(splitcomp);
	free(splitfreqs);
	free(splitpatterns);
	freequartets();
	free_cmatrix(biparts);
	free_cmatrix(Identif);
	free_cmatrix(Seqchar);
	free_ivector(freeID);
	free_ivector(trueID);
}

/* compare function for quicksort - sort biggest numbers first */
int ulintcmp(const ulint *a, const ulint *b)
{
	if (*a>*b) return (-1);
	else if (*a<*b) return 1;
	else return 0;
}
